/*
 * libconfigfile - configfile.c
 *	version 0.0.1
 *
 * Copyright (C) 2005  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: configfile.c,v 1.3 2005/07/27 20:15:07 krzysiek Exp $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <configfile.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#ifdef CF_DEBUG
#define DEBUG(fmt...)	fprintf(stderr, fmt)
#else
#define DEBUG(fmt...)	do { } while (0)
#endif

#define ERROR(fmt...)	fprintf(stderr, fmt)

static struct cf_section *cf_section_new(struct cf_space *space,
		const char *name)
{
	struct cf_section *sect;

	sect = (struct cf_section *) malloc(sizeof(struct cf_section));
	if (sect == NULL)
		return NULL;
	sect->name = strdup(name);
	if (sect->name == NULL) {
		free(sect);
		return NULL;
	}
	sect->fe = NULL;
	sect->next = NULL;
	if (space->fs != NULL)
		space->ls->next = sect;
	else
		space->fs = sect;
	space->ls = sect;	
	return sect;
}

static struct cf_entry *cf_entry_new(struct cf_section *sect,
		const char *name, const char *value)
{
	struct cf_entry *entry;

	entry = (struct cf_entry *) malloc(sizeof(struct cf_entry));
	if (entry == NULL)
		return NULL;
	entry->name = strdup(name);
	if (entry->name == NULL) {
		free(entry);
		return NULL;
	}
	entry->value = strdup(value);
	if (entry->value == NULL) {		
		free(entry);
		return NULL;
	}	
	entry->next = NULL;
	if (sect->fe != NULL)
		sect->le->next = entry;
	else
		sect->fe = entry;
	sect->le = entry;
	return entry;
}

int cf_space_init(struct cf_space *space)
{
	space->fs = NULL;
	return 0;
}

int cf_space_lock(struct cf_space *space)
{
	DEBUG("cf_space_lock(): Operation not supported\n");
	return -ENOSYS;
}

int cf_space_unlock(struct cf_space *space)
{
	DEBUG("cf_space_unlock(): Operation not supported\n");
	return -ENOSYS;
}

struct cf_section *cf_section_get(struct cf_space *space, const char *name)
{
	struct cf_section *tmp;

	for (tmp = space->fs; tmp != NULL; tmp = tmp->next)
		if (strcmp(tmp->name, name) == 0)
			return tmp;
	return NULL;
}

struct cf_entry *cf_entry_get(struct cf_section *section, const char *name)
{
	struct cf_entry *tmp;

	for (tmp = section->fe; tmp != NULL; tmp = tmp->next)
		if (strcmp(tmp->name, name) == 0)
			return tmp;
	return NULL;
}

const char *cf_entry_get_value(struct cf_section *section, const char *name)
{
	struct cf_entry *tmp;

	for (tmp = section->fe; tmp != NULL; tmp = tmp->next)
		if (strcmp(tmp->name, name) == 0)
			return tmp->value;
	return NULL;
}

const char *cf_get_value(struct cf_space *space, const char *section_name,
		const char *entry_name)
{
	struct cf_section *sect;

	sect = cf_section_get(space, section_name);
	if (sect != NULL)
		return cf_entry_get_value(sect, entry_name);
	return NULL;
}

int cf_space_loadfile(struct cf_space *space, const char *fname)
{
	FILE *f;
	struct cf_section *sect = NULL;
	int c;
	int ret = 0;
	char buff[256], buff2[256];

	if ((f = fopen(fname, "r")) == NULL)
		return -errno;
	while ((c = getc(f)) != EOF) {
		if (c == '#') {
			do {
				c = getc(f);
				if (c == EOF)
					break;
			} while (c != '\n');
			continue;
		}
		if ((c == ' ') || (c == '\t') || (c == '\n'))
			continue;
		if (c == '[') {			
			fscanf(f, "%256[^]]", buff);
			DEBUG("libconfigfile: adding section: %s\n", buff);
			c = getc(f);
			if (c != ']') {
				ret = -EINVAL;
				DEBUG("libconfigfile: parse error.\n");
				break;
			}
			sect = cf_section_new(space, buff);
			continue;
		}
		if (sect == NULL) {
			ERROR("libconfigfile: No section defined yet.\n");
			break;
		}
		ungetc(c, f);
		fscanf(f, "%256[0-9a-zA-z]", buff);
		do {
			c = getc(f);
			if (c == EOF) {
				DEBUG("warning: EOF in entry line\n");
				break;
			}
			if (c == '\n') {
				DEBUG("warning: new line in entry\n");
				break;
			}
		} while (c != '=');
		fscanf(f, "%256[^\n]", buff2);
		cf_entry_new(sect, buff, buff2);
	}
	fclose(f);
	return ret;
}

static int cf_section_clear(struct cf_section *sect)
{
	struct cf_entry *tmp;

	for (; sect->fe != NULL;) {
		tmp = sect->fe;
		sect->fe = sect->fe->next;
		free(tmp->value);
		free(tmp->name);
		free(tmp);
	}
	return 0;
}

static int cf_space_clear(struct cf_space *space)
{
	struct cf_section *tmp;

	for (; space->fs != NULL;) {
		tmp = space->fs;
		space->fs = space->fs->next;
		cf_section_clear(tmp);
		free(tmp);
	}
	return 0;
}

int cf_section_mget_init(struct cf_section_mget_state *state,
		struct cf_space *space, const char *name)
{
	state->sect = cf_section_get(space, name);
	return 0;
}

struct cf_section *cf_section_mget_get(struct cf_section_mget_state *state)
{
	struct cf_section *tmp;

	tmp = state->sect;
	if (state->sect != NULL) {
		state->sect = state->sect->next;
		while ((state->sect != NULL) &&
				(strcmp(tmp->name, state->sect->name)))
			state->sect = state->sect->next;
	}
	return tmp;
}

int cf_section_mget_done(struct cf_section_mget_state *state)
{
	return 0;
}

int cf_entry_mget_init(struct cf_entry_mget_state *state,
		struct cf_section *sect, const char *name)
{
	state->entry = cf_entry_get(sect, name);
	return 0;
}

struct cf_entry *cf_entry_mget_get(struct cf_entry_mget_state *state)
{
	struct cf_entry *tmp;

	tmp = state->entry;
	if (state->entry != NULL) {
		state->entry = state->entry->next;
		while ((state->entry != NULL) &&
				(strcmp(tmp->name, state->entry->name)))
			state->entry = state->entry->next;
	}
	return tmp;
}

const char *cf_entry_mget_get_value(struct cf_entry_mget_state *state)
{
	struct cf_entry *tmp;

	tmp = state->entry;
	if (state->entry != NULL) {
		state->entry = state->entry->next;
		while ((state->entry != NULL) &&
				(strcmp(tmp->name, state->entry->name)))
			state->entry = state->entry->next;
	}
	return (tmp != NULL) ? tmp->value : NULL;
}

int cf_entry_mget_done(struct cf_entry_mget_state *state)
{
	return 0;
}

int cf_space_done(struct cf_space *space)
{	
	return cf_space_clear(space);
}
